// Mongoose configuration file. For full list of build options, see
// https://mongoose.ws/docs/getting-started/build-options/

// DO NOT EDIT THIS FILE, BECAUSE CODE REGENERATION WILL DESTROY
// YOUR EDITS. THIS FILE INCLUDES THIS PROJECT'S main.h, SO
// INSTEAD, ADD BUILD OPTIONS TO Core/Inc/main.h

#pragma once

#define MG_ARCH MG_ARCH_CUBE
